module.exports = {
  description: 'https://github.com/chartjs/Chart.js/issues/7734',
  config: {
    type: 'line',
    options: {
      scales: {
        y: {
          max: 1225.2,
          min: 369.5,
        },
        x: {
          display: false
        }
      }
    }
  },
  options: {
    spriteText: true
  }
};
