@require '~vuepress-plugin-tabs/dist/themes/default.styl'

.theme-default-content
  &:not(.custom)
    max-width: unset

.sidebar-group.is-sub-group.depth-1
  > .sidebar-group-items
    border-left 1px solid rgba($accentColor, 0.25)

  > .sidebar-heading:not(.open)
    border-left 1px solid rgba($accentColor, 0.25)
    margin-left: 0

  > .sidebar-heading
    padding-left calc(1.475rem - 1px)
    transition border-color .25s
    padding 0.35rem 1.475rem
    border-left-width 3px
    margin-left -1px
    font-size 1em
    line-height 1.4
    opacity 1 !important

    &.active, &.open
      border-left-color $accentColor
      color $accentColor
      font-weight bold

    >.arrow
      display none

    >.sidebar-group-items
      padding-left: 0

.sidebar-group.is-sub-group.depth-1:hover .sidebar-heading:not(.open)
  color $accentColor
  margin-left -1px
  border-left 3px solid rgba($accentColor, 0.25)
  padding-left calc(1.475rem - 1px)
